<?php
/*
 * ClientDemo.php
 *
 * LICENSE: This source file is subject to version 1.0 of the OpenFlyers license that is available through
 * the world-wide-web at the following URI: http://www.openflyers.com/license/semifreelicense1_0.txt.
 * If you did not receive a copy of the OpenFlyers License and are unable to obtain it through the web,
 * please send a note to contact@openflyers.com so we can mail you a copy immediately.
 *
 * @module      OAuth2
 * @package     OpenFlyers
 * @author      Guillaume Obin <gobin@enseirb-matmeca.fr>
 * @copyright   2008 OPENFLYERS S.A.R.L. <contact@openflyers.com>
 * @license     http://www.openflyers.com/license/semifreelicense1_0.txt  OpenFlyers License
 * @link        http://www.openflyers.com
 * @since       Mon, September 06, 2021
 *
 */

class ClientDemo
{
    /**
     * Parsed configuration file for the authorization code grant type
     * 
     * @var array
     */
    private $configAuth = null;

    /**
     * Parsed configuration file for the client credentials grant type
     * 
     * @var array
     */
    private $configCred = null;

    /**
     * Demo client page URL
     * 
     * @var string
     */
    private $baseURL    = null;

    /**
     * Error string to display
     * 
     * @var Exception
     */
    private $error      = null;

    /**
     * addLog
     * 
     * Add an entry in the journal table and the matching
     * entries in the log table
     * @param   array       $logData        Data to push into the log file.
     * @return  boolean
     */
    static public function addLog($logData)
    {
        $headerData     = array(date('Y-m-d H:i:s'), 'openflyers.com', 'oauth2-demo' /* Where does the error appear ? */);
        $headerString   = implode(';', $headerData);
        $logString      = implode(';', $logData);

        $file   = fopen('/var/logError/oauth2-demo.log', 'a+');
        $status = fwrite($file, $headerString . ' : ' . $logString . "\n");
        fclose($file);

        return ($status) ? true : false;
    }

    /**
     * Save a variable named $name with the value $value in the session registry.
     *
     * @param string $name input variable name
     * @param mixed $value input variable value
     */
    static public function addValue($name, $value)
    {
        $_SESSION[$name] = $value;
    }

    /**
     * Get a value of a variable stored in the session by addValue
     *
     * @param string $name
     * @param mixed $default
     * @return mixed
     */
    static public function getFromSession($name, $default = null)
    {
        return isset($_SESSION[$name]) ? $_SESSION[$name] : $default;
    }

    /**
     * Delete a variable stored in the SESSION
     *
     * @param string $name name of the variable to remove
     */
    static public function kill($name)
    {
        if (isset($_SESSION[$name])) {
            unset($_SESSION[$name]);
        }
    }

    /**
     * Constructor
     * 
     * Initializes the base URL and the config arrays
     */
    public function __construct()
    {
        $this->setBaseUrl();
        $this->setConfig();
    }

    /**
     * addError
     *
     * @param   Exception   $error  Error to display after the menu.
     */
    public function addError($error)
    {
        $this->error = $error;
    }

    /**
     * This function is used to make API calls to the authorization server or the resource server
     *
     * @param   string      $url        Endpoint URL to the authorization server or the resource server, depending on the use case.
     * @param   array|null  $post       POST parameters array to use for the POST request only.
     * @param   array|null  $token      Full token in array format containing the token typen, the access token and/or the refresh token.
     * @param   bool        $auth       Flag to indicate if the URL is the authorization endpoint or not.
     * @param   string      $configType Indicates the config array to use. Can be either 'configAuth' or 'configCred'.
     * @param   array       $headers    Headers to add to the request if any.
     *
     * @return  string|bool             Response of the cURL request.
     * @throws  Exception
     */
    public function apiRequest($url, $post = null, $token = null, $auth = true, $configType = 'configAuth', $headers = array())
    {
        // Getting the configuration file according to the configuration type chosen
        $configFile = null;

        if ($configType == 'configAuth') {
            // Using authorization code config file
            $configFile = $this->configAuth;
        } else if ($configType == 'configCred') {
            // Using client credentials config file
            $configFile = $this->configCred;
        }

        // Initializing cURL object for the request
        $ch = curl_init($url);

        // Setting up useful cURL flags
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($ch, CURLOPT_HEADER, true);

        // Defining the default HTTP method to use
        $method = 'get';

        // If there is content in the $post variable
        if ($post) {
            // Setting up the HTTP method to POST
            $method   = 'post';

            // Building POST query for the request
            $postData = http_build_query($post);

            // Setting up correct flags for a POST request
            curl_setopt($ch, CURLOPT_POST, true);
            curl_setopt($ch, CURLOPT_POSTFIELDS, $postData);

            // Setting up headers for signing (HTTP-Signature)
            $headersToSign['Content-Type'] = 'application/x-www-form-urlencoded';
            $headersToSign['digest']       = 'SHA-256=' . base64_encode(hash('sha256', $postData, true));
        }

        // Getting parsed URL from the cURL object
        $urlComponents = parse_url($url);

        // Defining default headers to sign for HTTP-Signature
        $headersToSign['(request-target)'] = $method . ' ' . $urlComponents['path'];
        $headersToSign['Host']             = $urlComponents['host'];
        $headersToSign['Date']             = gmdate('D, j M Y H:i:s T');

        // Generating the signature header
        $keyId                 = openssl_x509_fingerprint(file_get_contents($configFile['sign_cert']));
        $privateKey            = file_get_contents($configFile['sign_key']);
        $headers['Signature']  = $this->generateSignatureHeader($headersToSign, $keyId, $privateKey);

        // Defining remaining headers to complete the request
        $headers['Accept']     = 'application/json';
        $headers['User-Agent'] = $this->baseURL;
        unset($headersToSign['(request-target)']);
        $headers               += $headersToSign;

        // If there is a token set in parameters, adding authorization header
        if ($token) {
            $headers['Authorization'] = $token['token_type'] . ' ' . $token['access_token'];
        }

        // Formatting the headers to HTTP standards
        $httpFormattedHeaders = [];
        foreach ($headers as $key => $value) {
            $httpFormattedHeaders[] = trim($key) . ': ' . trim($value);
        }

        // Setting up HTTP headers in the cURL request
        curl_setopt($ch, CURLOPT_HTTPHEADER, $httpFormattedHeaders);
        curl_setopt($ch, CURLINFO_HEADER_OUT, true);

        // for development environment only
        // curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
        // curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
        // curl_setopt($ch, CURLOPT_VERBOSE, true);

        // Defining TLS client information with the TLS version, certificate and private key for mTLS
        curl_setopt($ch, CURLOPT_SSLVERSION, CURL_SSLVERSION_TLSv1_2);
        curl_setopt($ch, CURLOPT_CAINFO, $configFile['auth_cacert']);
        curl_setopt($ch, CURLOPT_SSLCERT, $configFile['auth_cert']);
        curl_setopt($ch, CURLOPT_SSLKEY, $configFile['auth_key']);

        // Executing the cURL request
        $response = curl_exec($ch);

        // Closing the cURL connection to free up space
        curl_close($ch);

        // In case an authentication request is executed
        if ($auth) {
            // Separating headers from body
            list($firstResponseHeaders, $firstResponseBody) = explode("\r\n\r\n", $response, 2);

            /**
             * Sometimes, there are two headers, the first one says '100 Continue' and the second one
             * is the one of interest. The issue is that the second headers are in the response body,
             * so there is a second required separation to get the correct headers and the correct body.
             */
            if (preg_match('/100/', $firstResponseHeaders)) {
                list($secondResponseHeaders, $responseBody) = explode("\r\n\r\n", $firstResponseBody, 2);
            } else {
                $secondResponseHeaders = $firstResponseHeaders;
                $responseBody          = $firstResponseBody;
            }

            // Defining digest and signature variables to get them dumped from the headers
            $responseHeadersDigest = '';
            $responseHeadersSignature = '';

            // Converting headers from string to array
            $responseHeadersArray = explode("\r\n", $secondResponseHeaders);
            $responseProtocol = array_shift($responseHeadersArray);

            // In case the response is not '200 OK', stopping the execution and throwing an error
            if (!preg_match('/200/', $responseProtocol)) {
                static::addLog(array('ERROR_WITH_OAUTH2_DEMO_AUTH_PROCESS', $firstResponseHeaders, $firstResponseBody));
                throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
            }

            // Storing headers in an array with the values associated to each key
            foreach ($responseHeadersArray as $value) {
                list($responseHeadersKeys, $responseHeadersValue) = explode(": ", $value, 2);
                $responseHeaders[strtolower($responseHeadersKeys)] = $responseHeadersValue;
            }

            // Defining digest algorithm and key to retrieve them from the digest header
            $responseDigestAlgo = '';
            $responseDigestKey = '';

            // Extracting digest and signature headers
            foreach ($responseHeaders as $key => $value) {
                if ($key === "digest") {
                    // Storing the digest in a variable
                    $responseHeadersDigest = $value;

                    // Stripping SHA algorithm for later comparison
                    list($responseDigestAlgo, $responseDigestKey) = explode("=", $responseHeadersDigest, 2);
                } else if ($key === "signature") {
                    // Storing the signature in a variable
                    $responseHeadersSignature = $value;
                }
            }

            // Calculating response digest
            $responseDigest = base64_encode(hash(strtolower(str_replace("-", "", $responseDigestAlgo)), $responseBody, true));

            // Checking digest validity
            if ($responseDigest !== $responseDigestKey) {
                // Logging and throwing the error because this means there has been a tamper
                static::addLog(array(
                    'ERROR_WITH_OAUTH2_DEMO_AUTH_PROCESS',
                    'There was an error while comparing digests. Digests are not the same!',
                    'Expected digest: ' . $responseDigestKey,
                    'Calculated digest: ' . $responseDigest
                ));
                throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
            }

            // Extracting variables from signature header
            $signatureArray = explode(",", $responseHeadersSignature);
            foreach ($signatureArray as $value) {
                list($signatureKey, $signatureValue) = explode("=", $value, 2);
                $signature[$signatureKey] = trim($signatureValue, '"');
            }

            // Generating siging string
            $signingStringArray = explode(" ", $signature['headers']);
            $signingString = '';
            foreach ($signingStringArray as $value) {
                $signingString = $signingString . trim($value) . ": " . trim($responseHeaders[$value]) . "\n";
            }

            // Trimming last '\n' character
            $signingString = trim($signingString);

            // Decoding signature from base64
            $decodedSignature = base64_decode($signature['signature']);

            // Verifying signature
            $signatureVerify = openssl_verify($signingString, $decodedSignature, openssl_get_publickey(file_get_contents($configFile['sign_cert_server'])), 'RSA-SHA256');

            // Checking the signature validity
            if (!$signatureVerify) {
                // Defining the error array to fill
                $errorArray = array(
                    'ERROR_WITH_OAUTH2_DEMO_AUTH_PROCESS',
                    'There was an error while verifying the signature. Signature is not correct!'
                );

                // Filling the error array with OpenSSL error strings
                while ($opensslError = openssl_error_string())
                    array_push($errorArray, $opensslError);

                // Logging and throwing the error because the signature cannot be invalid
                static::addLog($errorArray);
                throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
            }

            // Returning the response body
            return json_decode($responseBody, true);
        }

        // When the request is not an authentication request, then returning the full response without treatment
        return $response;
    }

    /**
     * This function initiate the login process for the authorization code grant type
     * 
     * @param   string  $buttonValue    Value of the submit button to behave differently regarding if the user is connected or not.
     * 
     * @return  void
     */
    public function authCodeLogin($butttonValue)
    {
        // If the user is not connected
        if (is_null(static::getFromSession('auth_token')) and $butttonValue == 'Se connecter') {
            // Proceeding with the login function to connect the user
            $this->login($this->configAuth['client_id'], $this->baseURL, $this->configAuth['authorize_uri']);
        } else {
            // revoke token
            $this->apiRequest(
                $this->configAuth['revoke_uri'],
                ['access_token' => static::getFromSession('auth_token')['access_token']],
                null,
                false
            );
            // If the user is already connected, killing the session
            static::kill('auth_token');
            static::kill('authResource');
        }
    }

    /**
     * This function initiate the login process of the client credentials grant type
     * 
     * @param   string      $buttonValue    Value of the submit button to behave differently regarding if the user is connected or not.
     * 
     * @return  void
     * @throws  Exception
     */
    public function clientCredLogin($butttonValue)
    {
        // If the user is not connected
        if (is_null(static::getFromSession('cred_token')) and $butttonValue == 'Se connecter') {
            // Retrieving the token from the authorization server
            $token = $this->apiRequest(
                $this->configCred['token_uri'],
                [
                    'grant_type'    => 'client_credentials',
                    'client_id'     => $this->configCred['client_id'],
                    'client_secret' => empty($this->configCred['client_secret']) ? null : $this->configCred['client_secret'],
                    'scope'         => 'genericreports.readonly reports.readonly'
                ],
                null,
                true,
                'configCred'
            );

            /**
             * Storing the access token in the session so the user is "connected"
             * only if the request has been successfully executed
             */
            if (isset($token['access_token'])) {
                // Storing the token in the user session
                static::addValue('cred_token', $token);

                // Redirecting the user's browser to the "home" page
                header('Location: ' . $this->baseURL);
            } else {
                // Logging and throwing an error because there is an issue while retrieving the token
                static::addLog(array('ERROR_WITH_OAUTH2_DEMO_CLIENT_CRED_TOKEN_RETRIEVAL', $token));
                throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
            }
        } else {
            // revoke token
            $this->apiRequest(
                $this->configCred['revoke_uri'],
                ['access_token' => static::getFromSession('cred_token')['access_token']],
                null,
                false,
                'configCred'
            );
            // If the user is already connected, killing the session
            static::kill('cred_token');
        }
    }

    /**
     * This function displays the demo page
     * 
     * @param   string  $rootPath   Defining the document root path in case the demo code is in a subdirectory.
     * 
     * @return  void
     */
    public function displayDemoPage()
    {
        $this->xhtmlHeader();
        $this->xhtmlBody();
        $this->xhtmlError();
        $this->xhtmlForm();
        $this->xhtmlFooter();
    }

    /**
     * This function retrieves the user identifier from the resource server for the authorization code grant type
     * 
     * @return  void
     * @throws  Exception
     */
    public function getAuthResource()
    {
        // Building the query array
        $repoQueryParameters = [
            'resource_type' => 'user_information',              // Resource type to retrieve
            'client_id'     => $this->configAuth['client_id']   // Client identifier for security measures
        ];

        // Executing the request to the resource server
        $resources = $this->apiRequest(
            $this->configAuth['resource_uri'],
            $repoQueryParameters,
            static::getFromSession('auth_token'),
            false
        );

        // Separating headers from body
        list($resourceHeader, $resourceBody) = explode("\r\n\r\n", $resources, 2);

        // Converting headers from string to array
        $resourceHeaders = explode("\r\n", $resourceHeader);
        $firstHeaderLine = array_shift($resourceHeaders);

        // Checking if the resource has been successfully retrieved or not
        if (!preg_match('/200/', $firstHeaderLine)) {
            // Logging and throwing an error because there is an issue while retrieving the resource
            static::addLog(array('ERROR_WITH_OAUTH2_DEMO_CLIENT_AUTH_RESOURCE_RETRIEVAL', $resourceBody));
            throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
        }

        // Adding the user information in the user session
        static::addValue('authResource', json_decode($resourceBody, true));
    }

    /**
     * This function is used to convert an authorization code to an access token
     * 
     * @return  void
     * @throws  Exception
     */
    public function getAuthToken()
    {
        // Retrieving state value from GET and COOKIE for comparison purposes
        $getState    = filter_input(INPUT_GET, 'state');
        $cookieState = filter_input(INPUT_COOKIE, 'state');

        /**
         * The authorization server MUST redirect the browser here with the exact same state parameter sent before,
         * so checking if it is indeed the same to detect if the OAuth flow has been tampered with or not
         */
        if (!$getState || $cookieState != $getState) {
            // Logging and throwing an error because there is an issue while retrieving the token
            static::addLog(array('ERROR_WITH_OAUTH2_DEMO_AUTH_CODE_TOKEN_RETRIEVAL', 'Values for state variable are different'));
            throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
        }

        // Getting code and code_verifier from GET and COOKIE to initiate the request
        $getCode            = filter_input(INPUT_GET, 'code');
        $cookieCodeVerifier = filter_input(INPUT_COOKIE, 'code_verifier');

        /**
         * Communicating directly with the authorization server to exchange the code received against an access token,
         * the id/secret pair is sent to authenticate the client, the redirect_uri is sent to verify the code validity
         */
        $token = $this->apiRequest(
            $this->configAuth['token_uri'],
            [
                'grant_type'    => 'authorization_code',
                'client_id'     => $this->configAuth['client_id'],
                'client_secret' => empty($this->configAuth['client_secret']) ? null : $this->configAuth['client_secret'],
                'redirect_uri'  => $this->baseURL,
                'code'          => $getCode,
                'code_verifier' => $cookieCodeVerifier
            ]
        );

        // Resetting the cookies to remove state and code_verifier for security reasons
        setcookie('state',         '', time() - 60, '/');
        setcookie('code_verifier', '', time() - 60, '/');

        /**
         * Storing the access token in the session so the user is "connected"
         * only if the request has been successfully executed
         */
        if (isset($token['access_token'])) {
            // Storing the token in the user session
            static::addValue('auth_token', $token);

            // Redirecting the user's browser to the "home" page
            header('Location: ' . $this->baseURL);
        } else {
            // Logging and throwing an error because there is an issue while retrieving the token
            static::addLog(array('ERROR_WITH_OAUTH2_DEMO_AUTH_CODE_TOKEN_RETRIEVAL', $token));
            throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
        }
    }

    /**
     * This function is used to get a report from the type given in parameter
     * 
     * @param   string      $reportType Report type to retrieve from the resource server
     * 
     * @return  never
     * @throws  Exception
     */
    public function getReportResource($reportType)
    {
        // Defining the report ID regarding the report type chosen
        $reportId = null;

        /**
         * The replacement list is empty because the demo relies on the default values to retrieve the report
         * except for the custom reports that might need values to be set regarding of the SQL query set up
         * @see below
         */
        $replacementList = [];

        if ($reportType == 'generic_report') {
            // Report ID for generic report
            $reportId        = 137;
        } else if ($reportType == 'report') {
            // Report ID for custom report
            $reportId = 135;
        }

        /**
         * Here we assign default values in the replacement list to be set regarding of the SQL query set up (for the two reports 137 and 135)
         * In order to avoid 500 Internal Server Error due to PDOException
         * It is advisable to handle situations where default values aren't exist in the database
         * @link https://openflyers.com/fr/doc/of4/API-OpenFlyers#500-Internal-Server-Error-en-récupérant-le-rapport
         */
        $replacementList = ['year' => 2023];

        // Building query array
        $repoQueryParameters = [
            'resource_type' => 'generic_report',
            'client_id' => $this->configCred['client_id'],
            'report_id' => $reportId,
            'replacementList' => $replacementList
        ];

        // Making a call to the resource server
        $resources = $this->apiRequest(
            $this->configCred['resource_uri'],
            $repoQueryParameters,
            static::getFromSession('cred_token'),
            false,
            'configCred'
        );

        // Separating headers from body
        list($resourceHeader, $resourceBody) = explode("\r\n\r\n", $resources, 2);
        $resourceHeaders = explode("\r\n", $resourceHeader);
        $responseCode    = array_shift($resourceHeaders);

        // Checking the status of the request, if it is OK then displaying the content, otherwise logging and throwing the error
        if (preg_match('/200/', $responseCode)) {
            // Dumping headers to insert them in current page
            foreach ($resourceHeaders as $key => $value) {
                /**
                 * Avoid Set-Cookie header to avoid session issues
                 * To prevent download issues with files, avoid using the "Transfer-Encoding: chunked" header
                 * The header "chunked" is produced in the production environment by the Nginx server
                 */
                if (!preg_match('/Set-Cookie/', $value) && !preg_match('/Transfer-Encoding: chunked/', $value))
                    header($value);
            }

            // Displaying the content to download
            echo json_decode($resourceBody);

            // Killing the script to only download CSV content
            die();
        } else {
            // Logging and throwing the error
            static::addLog(array('ERROR_WITH_OAUTH2_DEMO', $resourceBody));
            throw new Exception('OAUTH_ERROR_OCCURED_DEMO');
        }
    }

    /**
     * This function takes a code_verifier and outputs the corresponding code_challenge
     *
     * @param string $codeVerifier the generated code_verifier
     *
     * @return string the computed code_challenge
     */
    private function computeCodeChallenge(string $codeVerifier): string
    {
        return strtr(rtrim(base64_encode(hash('sha256', $codeVerifier, true)), '='), '+/', '-_');
    }

    /**
     * This function takes an optional string length and outputs a random code_verifier string
     *
     * @param int $length the length of the output code_verifier. Default = 128
     *
     * @return string the code_verifier
     * @throws Exception
     */
    private function generateCodeVerifier($length = 128): string
    {
        $characters = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ-._~';
        $outputCode = '';

        for ($i = 0; $i < $length; $i++) {
            $index      = random_int(0, strlen($characters) - 1);
            $outputCode .= $characters[$index];
        }

        return $outputCode;
    }

    /**
     * This function generates the full signature header line from a set of headers, a certificate and the linked private key
     *
     * @param array  $headersToSign the headers to sign, in the $key => $value format
     * @param string $certificate   the certificate linked to the used private key
     * @param string $privateKey    the private key used to sign the headers
     *
     * @return string the full signature header line
     */
    private function generateSignatureHeader(array $headersToSign, string $certificate, string $privateKey): string
    {
        // generating the signing string and header list
        $headers         = '';
        $signatureString = '';
        foreach ($headersToSign as $key => $value) {
            $normalizedHeaderKey = trim(strtolower($key));
            $headers             .= $normalizedHeaderKey . ' ';
            $signatureString     .= $normalizedHeaderKey . ': ' . trim($value) . "\n";
        }
        $headers         = trim($headers);
        $signatureString = trim($signatureString);

        // signing the signing string
        $signature = '';
        openssl_sign($signatureString, $signature, openssl_get_privatekey($privateKey), 'RSA-SHA256');
        $signature = base64_encode($signature);

        // compiling the header line
        return "keyId=\"$certificate\",algorithm=\"rsa-sha256\",headers=\"$headers\",signature=\"$signature\"";
    }

    /**
     * This function is used to initiate the authentication code flow.
     *
     * @param   string      $clientID       Client identifier to use for the request.
     * @param   string      $redirectURL    URL to use for the redirection after the authentication process.
     * @param   string      $authorizeURL   Endpoint URL of the authorization server.
     *
     * @return  void
     * @throws  Exception
     */
    private function login(string $clientID, string $redirectURL, string $authorizeURL): void
    {
        // This unguessable string is used to prevent csrf attacks
        $state         = bin2hex(random_bytes(16));

        // Generate code_verifier and code_challenge for PKCE    
        $codeVerifier  = $this->generateCodeVerifier();
        $codeChallenge = $this->computeCodeChallenge($codeVerifier);

        /**
         * Required parameters for the redirection, redirect_uri is where the browser should be redirected
         * when the user grants (or denies) access, scope are the authorizations (rights) requested
         */
        $params = [
            'response_type'         => 'code',
            'client_id'             => $clientID,
            'redirect_uri'          => $redirectURL,
            'scope'                 => 'default.login',
            'state'                 => $state,
            'code_challenge'        => $codeChallenge,
            'code_challenge_method' => 'S256'
        ];

        /**
         * Setting up state and code_verifier values in cookie rather than session because there is an issue with
         * the user session being destroyed when the login form is displayed resulting in the values being wiped out
         */
        setcookie('state',         $state,        time() + 600, '/');
        setcookie('code_verifier', $codeVerifier, time() + 600, '/');

        // Redirecting the browser to the authorization endpoint to obtain the authorization code
        header('Location: ' . $authorizeURL . '?' . http_build_query($params));
    }

    /**
     * This function sets the base URL globally to make redirections easier
     * 
     * @return  void
     */
    private function setBaseUrl()
    {
        // Build the baseURL, accounting for protocol, port, name and endpoint. Used for redirection
        $protocol      = isset($_SERVER['HTTPS']) ? 'https://' : 'http://';
        $port          = ($protocol == 'https://' && $_SERVER['SERVER_PORT'] == 443)
            || ($protocol == 'http://' && $_SERVER['SERVER_PORT'] == 80) ? '' : ':' . $_SERVER['SERVER_PORT'];

        // Storing the URL in an object variable
        $this->baseURL = $protocol . $_SERVER['SERVER_NAME'] . $port . $_SERVER['PHP_SELF'];
    }

    /**
     * This function loads both authorization code and client credentials configuration files
     * 
     * @return  void
     */
    private function setConfig()
    {
        $this->configAuth = json_decode(file_get_contents('./ssl/AuthCodeDemo/config.authcode.json'), true);
        $this->configCred = json_decode(file_get_contents('./ssl/ClientCredDemo/config.clientcred.json'), true);
    }

    /**
     * This function displays the HTML contentBox opening tags
     * 
     * @return  void
     */
    private function xhtmlBody()
    {
        echo '<!-- contentBox -->
        <div class="contentBox">';
    }

    /**
     * This function displays the HTML error box
     * 
     * @return  void
     */
    private function xhtmlError()
    {
        if ($this->error) echo '<div class="alert-container">
        <div class="warningRed">Une erreur est survenue lors de l\'exécution de la démonstration. Merci de contacter l\'administrateur du site si l\'erreur persiste.</div>
        </div>';
    }

    /**
     * This function displays the HTML footer content
     * 
     * @return  void
     */
    private function xhtmlFooter()
    {
        echo '<!-- contentBox -->
    <div class="contentBox">
        
    </div><!-- /contentBox -->
        </body>
        </html>';
    }

    /**
     * This function displays the HTML form content
     * 
     * @return  void
     */
    private function xhtmlForm()
    {
        echo '<div class="structure-identity" id="loginClubLogo"><a class="logo-link" href="https://openflyers.com"><img
        class="logo" src="./img/ofBanner57.png" alt="OpenFlyers" title="OpenFlyers" /></a></h1>
        </div>
        <div class="login-form">
        <form id="authorize_form" method="post">
            <h2 class="title">Démonstration de client OAuth 2.0</h2>
            <div class="form-inner centered">
                <div class="col-container">
                    <div class="form-col">
                        <h2 class="title">Authorization Code</h2>
                        <p class="alignCenter"><b>État de la connexion</b><br>'
                    . (is_null(static::getFromSession('auth_token')) ? '<a style="color: red;">Déconnecté</a>' : '<a style="color: green;">Connecté</a>')
                    . '</p>'
                    . (is_null(static::getFromSession('authResource')) ? '' : '<p class="alignCenter"><b>Identifiant utilisateur</b> : ' . static::getFromSession('authResource')['user_id'] . '</p>')
                    . (is_null(static::getFromSession('auth_token')) ? '' : '<div class="form-button-container">
                        <div class="form-field"><input type="submit"
                                value="Récupérer les informations utilisateur" name="resourceUserId"
                                id="resourceUserId" class="button flat big standard"></div>
                    </div>')
                    . '<div class="form-button-container">
                            <div class="form-field"><input type="submit" value="'
                    . (is_null(static::getFromSession('auth_token')) ? 'Se connecter' : 'Se déconnecter')
                    . '" name="authCode"
                                    id="authCode" class="button flat big blue" /></div>
                        </div>
                    </div>
                    <div class="form-col">
                        <h2 class="title">Client Credentials</h2>
                        <p class="alignCenter"><b>État de la connexion</b><br>'
                    . (is_null(static::getFromSession('cred_token')) ? '<a style="color: red;">Déconnecté</a>' : '<a style="color: green;">Connecté</a>')
                    . '</p>'
                    . (is_null(static::getFromSession('cred_token')) ? '' : '<div class="form-field combobox">
                        <div class="inner">
                            <span class="label">Rapport à récupérer&nbsp;:</span>
                            <span class="formw"><select id="reportCombo" name="reportCombo">
                                    <option value="generic_report">Rapport générique de démonstration</option>
                                    <option value="report">Rapport personnalisé de démonstration</option>
                                </select> </span>
                        </div>
                    </div>
                    <div class="form-button-container">
                        <div class="form-field"><input type="submit" value="Récupérer le rapport"
                                name="resourceReport" id="resourceReport" class="button flat big standard">
                        </div>
                    </div>')
                    . '<div class="form-button-container">
                            <div class="form-field"><input type="submit" value="'
                    . (is_null(static::getFromSession('cred_token')) ? 'Se connecter' : 'Se déconnecter')
                    . '" name="clientCred"
                                    id="clientCred" class="button flat big blue" /></div>
                        </div>
                    </div>
                </div>
            </div>
        </form>
        </div>';
    }

    /**
     * This function displays the HTML header content
     * 
     * @return  void
     */
    private function xhtmlHeader()
    {
        echo '<!DOCTYPE html>
        <html class="preloader-page" lang="fr">
        
        <head>
            <title>OpenFlyers - Démonstration de client OAuth 2.0</title>
            <meta charset="utf-8">
            <link rel="shortcut icon" href="./favicon.ico">
            <link rel="apple-touch-icon" href="./img/home-screen-icon-192x192.png">
            <link rel="stylesheet" type="text/css" href="./css/OAuth2Demo.min.css" />
        </head>
        
        <body>
            <!-- top-bar -->
            <div class="top-bar">
                <!-- brand-->
                <div class="brand">
                    <a alt="OpenFlyers" class="logo" href="https://openflyers.com" target="_blank" title="OpenFlyers"></a>
                    <a class="of-web-storefront" href="https://openflyers.com" target="_blank">OpenFlyers<span
                            class="version">4.2</span></a>
                </div><!-- /brand-->
                <!-- social-->
                <ul class="social">
                    <li><a class="picto facebook" href="https://www.facebook.com/openflyers" target="_blank"></a></li>
                    <li><a class="picto twitter" href="https://twitter.com/OpenFlyers" target="_blank"></a></li>
                    <li><a class="picto youtube" href="https://www.youtube.com/user/OpenFlyersFr" target="_blank"></a></li>
                    <li><a class="picto actualites" href="https://openflyers.com/fr/actualites" target="_blank"></a></li>
                </ul><!-- /social-->
            </div><!-- /top-bar -->';
    }
}
